import pytest

from .. import BlackScholesButterflyLong, BlackScholesButterflyShort

# Test parameters
test_S = 55.0  # Asset price of 55
test_K1 = 40.0  # Strike price of 40
test_K2 = 50.0  # Strike price of 50
test_K3 = 60.0  # Strike price of 60
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesButterflyLong:
    def test_init(self):
        # Assert K1 < K2 < K3
        with pytest.raises(AssertionError):
            BlackScholesButterflyLong(
                S=test_S, K1=60, K2=50, K3=40, T=test_T, r=test_r, sigma=test_sigma
            )
        with pytest.raises(AssertionError):
            BlackScholesButterflyLong(
                S=test_S, K1=41, K2=40, K3=50, T=test_T, r=test_r, sigma=test_sigma
            )
        # Assert K2 - K1 = K3 - K2 (symmetry)
        with pytest.raises(AssertionError):
            BlackScholesButterflyLong(
                S=test_S, K1=41, K2=50, K3=60, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        butterfly = BlackScholesButterflyLong(
            test_S, test_K1, test_K2, test_K3, test_T, test_r, test_sigma
        )
        test_methods = list(butterfly.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Long (call) butterfly = Call1 - 2 * Call2 + Call3
        for attr in test_methods:
            assert (
                getattr(butterfly, attr)()
                == getattr(butterfly.call1, attr)()
                - 2 * getattr(butterfly.call2, attr)()
                + getattr(butterfly.call3, attr)()
            )


class TestBlackScholesButterflyShort:
    def test_init(self):
        # Assert K1 < K2 < K3
        with pytest.raises(AssertionError):
            BlackScholesButterflyShort(
                S=test_S, K1=60, K2=50, K3=40, T=test_T, r=test_r, sigma=test_sigma
            )
        with pytest.raises(AssertionError):
            BlackScholesButterflyShort(
                S=test_S, K1=41, K2=40, K3=50, T=test_T, r=test_r, sigma=test_sigma
            )
        # Assert K2 - K1 = K3 - K2 (symmetry)
        with pytest.raises(AssertionError):
            BlackScholesButterflyShort(
                S=test_S, K1=41, K2=50, K3=60, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        butterfly = BlackScholesButterflyShort(
            test_S, test_K1, test_K2, test_K3, test_T, test_r, test_sigma
        )
        test_methods = list(butterfly.put1.get_all_greeks().keys()) + [
            "price",
        ]
        # Short (put) butterfly = -Put1 + 2 * Put2 - Put3
        for attr in test_methods:
            assert (
                getattr(butterfly, attr)()
                == -getattr(butterfly.put1, attr)()
                + 2 * getattr(butterfly.put2, attr)()
                - getattr(butterfly.put3, attr)()
            )



from . import BlackScholesCall, BlackScholesPut
from .base import BlackScholesStructureBase


class BlackScholesBullSpread(BlackScholesStructureBase):
    """
    Create bull spread option structure. \n
    - Bull Spread -> Call(K1) - Call(K2)

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st call \n
    :param K2: Strike price for 2nd call \n
    It must hold that K1 < K2. \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """

    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 < K2
        ), f"""1st strike price should be smaller than 2nd.
        Got K1={K1}, which is not smaller than K2={K2}.
        """
        self.call1 = BlackScholesCall(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.call2 = BlackScholesCall(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from two call options into a bull spread. \n
        All greeks and prices are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to bull spread.
        """
        call1_attr = getattr(self.call1, attribute_name)
        call2_attr = getattr(self.call2, attribute_name)
        return call1_attr() - call2_attr()


class BlackScholesBearSpread(BlackScholesStructureBase):
    """
    Create bear spread option structure. \n
    - Bear Spread -> Put(K1) - Put(K2)

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st put \n
    :param K2: Strike price for 2nd put \n
    It must hold that K1 > K2. \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """
    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 > K2
        ), f"""1st strike price should be larger than 2nd.
        Got K1={K1}, which is not larger than K2={K2}.
        """
        self.put1 = BlackScholesPut(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.put2 = BlackScholesPut(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from two put options into a bear spread. \n
        All greeks and prices are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to bear spread.
        """
        put1_attr = getattr(self.put1, attribute_name)
        put2_attr = getattr(self.put2, attribute_name)
        return put1_attr() - put2_attr()


class BlackScholesCalendarCallSpread(BlackScholesStructureBase):
    """
    Create a calendar call spread option structure. \n
    Horizontal Calendar Call Spread -> K1 == K2 \n 
    Diagonal Calendar Call Spread -> K1 != K2 \n

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st call to buy. \n
    :param K2: Strike price for 2nd call to sell. \n
    :param T1: Time till expiration in years for 1st call. \n
    :param T2: Time till expiration in years for 2nd call. \n
    (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """
    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T1: float,
        T2: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            T1 > T2
        ), f"""1st time to maturity should be longer than 2nd.
        Got T1={T1}, which is not longer than T2={T2}.
        """
        self.call1 = BlackScholesCall(S=S, K=K1, T=T1, r=r, sigma=sigma, q=q)
        self.call2 = BlackScholesCall(S=S, K=K2, T=T2, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from two put options into a calendar spread. \n
        All greeks and prices are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to the calendar spread.
        """
        call1_attr = getattr(self.call1, attribute_name)
        call2_attr = getattr(self.call2, attribute_name)
        return call1_attr() - call2_attr()
    
    
class BlackScholesCalendarPutSpread(BlackScholesStructureBase):
    """
    Create a calendar put spread option structure. \n
    Horizontal Calendar Put Spread -> K1 == K2 \n 
    Diagonal Calendar Put Spread -> K1 != K2 \n

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st put to buy. \n
    :param K2: Strike price for 2nd put to sell. \n
    :param T1: Time till expiration in years for 1st put. \n
    :param T2: Time till expiration in years for 2nd put. \n
    (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """
    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T1: float,
        T2: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            T1 > T2
        ), f"""1st time to maturity should be longer than 2nd.
        Got T1={T1}, which is not longer than T2={T2}.
        """
        self.put1 = BlackScholesPut(S=S, K=K1, T=T1, r=r, sigma=sigma, q=q)
        self.put2 = BlackScholesPut(S=S, K=K2, T=T2, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from two put options into a calendar spread. \n
        All greeks and prices are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to the calendar spread.
        """
        put1_attr = getattr(self.put1, attribute_name)
        put2_attr = getattr(self.put2, attribute_name)
        return put1_attr() - put2_attr()


from abc import ABC, abstractmethod
from math import erf, exp, log, pi, sqrt
from typing import Dict


class StandardNormalMixin:
    """
    Fast PDF and CDF calculations for standard normal distribution.
    """

    @staticmethod
    def _pdf(x: float) -> float:
        """PDF of standard normal distribution."""
        return exp(-(x**2) / 2.0) / sqrt(2.0 * pi)

    @staticmethod
    def _cdf(x):
        """CDF of standard normal distribution."""
        return (1.0 + erf(x / sqrt(2.0))) / 2.0


class BlackScholesBase(ABC, StandardNormalMixin):
    """
    Base functionality to calculate (European) prices
    and Greeks with the Black-Scholes-Merton formula.

    :param S: Price of underlying asset \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(self, S: float, K: float, T: float, r: float, sigma: float, q: float):
        # Parameters checks
        assert S > 0.0, f"Asset price (S) needs to be larger than 0. Got '{S}'"
        assert K > 0.0, f"Strike price (K) needs to be larger than 0. Got '{K}'"
        assert T > 0.0, f"Time to maturity (T) needs to be larger than 0. Got '{T}'"
        assert (
            sigma > 0.0
        ), f"Volatility (sigma) needs to be larger than 0. Got '{sigma}'"
        assert q >= 0.0, f"Annual dividend yield (q) cannot be negative. Got '{q}'"
        self.S, self.K, self.T, self.r, self.sigma, self.q = S, K, T, r, sigma, q

    @abstractmethod
    def price(self) -> float:
        """Fair value for option."""
        ...

    @abstractmethod
    def in_the_money(self) -> float:
        """Naive probability that option will be in the money at maturity."""
        ...

    @abstractmethod
    def delta(self) -> float:
        """Rate of change in option price
        with respect to the forward price (1st derivative).
        """
        ...

    @abstractmethod
    def spot_delta(self) -> float:
        """
        Delta discounted for interest rates.
        For the forward delta, use `delta`.
        """
        ...

    @abstractmethod
    def dual_delta(self) -> float:
        """1st derivative of option price with respect to the strike price."""
        ...

    def gamma(self) -> float:
        """
        Rate of change in delta with respect to the underlying asset price (2nd derivative).
        """
        return (
            exp(-self.q * self.T)
            * self._pdf(self._d1)
            / (self.S * self.sigma * sqrt(self.T))
        )

    def dual_gamma(self) -> float:
        """
        Rate of change in delta with respect to the strike price (2nd derivative).
        """
        return (
            exp(-self.r * self.T)
            * self._pdf(self._d2)
            / (self.K * self.sigma * sqrt(self.T))
        )

    def vega(self) -> float:
        """
        Rate of change in option price with respect to the volatility of the asset.
        """
        return self.S * self._pdf(self._d1) * sqrt(self.T)

    @abstractmethod
    def theta(self) -> float:
        """
        Rate of change in option price
        with respect to time (i.e. time decay).
        """
        ...

    @abstractmethod
    def epsilon(self) -> float:
        """Change in option price with respect to underlying dividend yield. \n
        Also known as psi."""
        ...

    @abstractmethod
    def rho(self) -> float:
        """Rate of change in option price
        with respect to the risk-free rate.
        """
        ...

    def lambda_greek(self) -> float:
        """Percentage change in option value per %
        change in asset price. Also called gearing.
        """
        return self.delta() * self.S / self.price()

    def vanna(self) -> float:
        """Sensitivity of delta with respect to change in volatility."""
        return -self._pdf(self._d1) * self._d2 / self.sigma

    @abstractmethod
    def charm(self) -> float:
        """Rate of change of delta over time (also known as delta decay)."""
        ...

    def vomma(self) -> float:
        """2nd order sensitivity to volatility."""
        return self.vega() * self._d1 * self._d2 / self.sigma

    def veta(self) -> float:
        """Rate of change in `vega` with respect to time."""
        return (
            -self.S
            * exp(-self.q * self.T)
            * self._pdf(self._d1)
            * sqrt(self.T)
            * (
                self.q
                + (self.r - self.q) * self._d1 / (self.sigma * sqrt(self.T))
                - (1.0 + self._d1 * self._d2) / (2.0 * self.T)
            )
        )

    def phi(self) -> float:
        """2nd order partial derivative with respect to strike price. \n
        Phi is used in the Breeden-Litzenberger formula. \n
        Breeden-Litzenberger uses quoted option prices
        to estimate risk-neutral probabilities.
        """
        sigma2 = self.sigma**2
        exp_factor = (
            -1.0
            / (2.0 * sigma2 * self.T)
            * (log(self.K / self.S) - ((self.r - self.q) - 0.5 * sigma2) * self.T) ** 2
        )
        return (
            exp(-self.r * self.T)
            * (1.0 / self.K)
            * (1.0 / sqrt(2.0 * pi * sigma2 * self.T))
            * exp(exp_factor)
        )

    def speed(self) -> float:
        """Rate of change in Gamma with respect to change in the underlying price."""
        return -self.gamma() / self.S * (self._d1 / (self.sigma * sqrt(self.T)) + 1.0)

    def zomma(self) -> float:
        """Rate of change of gamma with respect to changes in volatility."""
        return self.gamma() * ((self._d1 * self._d2 - 1.0) / self.sigma)

    def color(self) -> float:
        """Rate of change of gamma over time."""
        return (
            -exp(-self.q * self.T)
            * self._pdf(self._d1)
            / (2.0 * self.S * self.T * self.sigma * sqrt(self.T))
            * (
                2.0 * self.q * self.T
                + 1.0
                + (
                    2.0 * (self.r - self.q) * self.T
                    - self._d2 * self.sigma * sqrt(self.T)
                )
                / (self.sigma * sqrt(self.T))
                * self._d1
            )
        )

    def ultima(self) -> float:
        """Sensitivity of vomma with respect to change in volatility. \n
        3rd order derivative of option value to volatility.
        """
        d1d2 = self._d1 * self._d2
        return (
            -self.vega()
            / self.sigma**2
            * (d1d2 * (1.0 - d1d2) + self._d1**2 + self._d2**2)
        )

    def alpha(self) -> float:
        """Theta to gamma ratio. Also called "gamma rent".
        More info: "Dynamic Hedging" by Nassim Taleb, p. 178-181.
        """
        return abs(self.theta()) / (self.gamma() + 1e-9)

    def get_core_greeks(self) -> Dict[str, float]:
        """
        Get the top 5 most well known Greeks.
        1. Delta
        2. Gamma
        3. Vega
        4. Theta
        5. Rho
        """
        return {
            "delta": self.delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "rho": self.rho(),
        }

    def get_itm_proxies(self) -> Dict[str, float]:
        """Get multiple ways of calculating probability
        of option being in the money.
        """
        return {"in_the_money": self.in_the_money(), "dual_delta": self.dual_delta()}

    def get_all_greeks(self) -> Dict[str, float]:
        """Retrieve all Greeks for the Black-Scholes-Merton model
        implemented as a dictionary."""
        return {
            "delta": self.delta(),
            "spot_delta": self.spot_delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "epsilon": self.epsilon(),
            "rho": self.rho(),
            "lambda_greek": self.lambda_greek(),
            "vanna": self.vanna(),
            "charm": self.charm(),
            "vomma": self.vomma(),
            "veta": self.veta(),
            "phi": self.phi(),
            "speed": self.speed(),
            "zomma": self.zomma(),
            "color": self.color(),
            "ultima": self.ultima(),
            "dual_delta": self.dual_delta(),
            "dual_gamma": self.dual_gamma(),
            "alpha": self.alpha(),
        }

    @property
    def _d1(self) -> float:
        """1st probability factor that acts as a multiplication factor for stock prices."""
        return (1.0 / (self.sigma * sqrt(self.T))) * (
            log(self.S / self.K) + (self.r - self.q + 0.5 * self.sigma**2) * self.T
        )

    @property
    def _d2(self) -> float:
        """2nd probability parameter that acts as a multiplication factor for discounting."""
        return self._d1 - self.sigma * sqrt(self.T)


class Black76Base(ABC, StandardNormalMixin):
    """
    Base functionality to calculate (European) prices
    and Greeks with the Black-76 formula. \n
    This variant of the Black-Scholes-Merton model is
    often used for pricing options on futures and bonds.

    :param F: Futures price \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    """

    def __init__(self, F: float, K: float, T: float, r: float, sigma: float):
        # Some parameters must be positive
        for param in [F, K, T, sigma]:
            assert (
                param >= 0.0
            ), f"Some parameters cannot be negative. Got '{param}' as an argument."
        self.F, self.K, self.T, self.r, self.sigma = F, K, T, r, sigma

    @abstractmethod
    def price(self):
        """Fair value for option."""
        ...

    @abstractmethod
    def delta(self):
        """Rate of change in option price
        with respect to the futures price (1st derivative)."""
        ...

    def gamma(self) -> float:
        """
        Rate of change in delta with respect to the underlying stock price (2nd derivative).
        """
        return (
            exp(-self.r * self.T)
            * self._pdf(self._d1)
            / (self.F * self.sigma * sqrt(self.T))
        )

    def vega(self) -> float:
        """Rate of change in option price with respect to the volatility
        of underlying futures contract.
        """
        return self.F * exp(-self.r * self.T) * self._pdf(self._d1) * sqrt(self.T)

    @abstractmethod
    def theta(self) -> float:
        """
        Rate of change in option price
        with respect to time (i.e. time decay).
        """
        ...

    @abstractmethod
    def rho(self) -> float:
        """Rate of change in option price
        with respect to the risk-free rate.
        """
        ...

    def vanna(self) -> float:
        """Sensitivity of delta with respect to change in volatility."""
        return self.vega() / self.F * (1 - self._d1 / (self.sigma * sqrt(self.T)))

    def vomma(self) -> float:
        """2nd order sensitivity to volatility."""
        return self.vega() * self._d1 * self._d2 / self.sigma

    def alpha(self) -> float:
        """Theta to gamma ratio. Also called "gamma rent".
        More info: "Dynamic Hedging" by Nassim Taleb, p. 178-181.
        """
        return abs(self.theta()) / (self.gamma() + 1e-9)

    def get_core_greeks(self) -> Dict[str, float]:
        """
        Get the top 5 most well known Greeks.
        1. Delta
        2. Gamma
        3. Vega
        4. Theta
        5. Rho
        """
        return {
            "delta": self.delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "rho": self.rho(),
        }

    def get_all_greeks(self) -> Dict[str, float]:
        """Retrieve all Greeks for the Black76 model implemented as a dictionary."""
        return {
            "delta": self.delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "rho": self.rho(),
            "vanna": self.vanna(),
            "vomma": self.vomma(),
            "alpha": self.alpha(),
        }

    @property
    def _d1(self) -> float:
        """1st probability factor that acts as a multiplication factor for futures contracts."""
        return (log(self.F / self.K) + 0.5 * self.sigma**2 * self.T) / (
            self.sigma * sqrt(self.T)
        )

    @property
    def _d2(self) -> float:
        """2nd probability parameter that acts as a multiplication factor for discounting."""
        return self._d1 - self.sigma * sqrt(self.T)


class BlackScholesStructureBase(ABC):
    """
    Option structure base class. \n
    `_calc_attr` should be implemented for every option structure.
    """

    @abstractmethod
    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from several put and call options.

        Ex. Long Straddle \n
        ```python
        def _calc_attr(self, attribute_name: str) -> float:
            put_attr = getattr(self.put1, attribute_name)
            call_attr = getattr(self.call1, attribute_name)
            return put_attr() + call_attr()
        ```
        In this way all greeks and price are combined in the same way. \n
        In this case only simple addition is performed.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value (float)
        """
        ...

    def price(self) -> float:
        """Fair value of Black-Scholes option structure."""
        return self._calc_attr(attribute_name="price")

    def delta(self) -> float:
        """Rate of change in structure price
        with respect to the asset price (1st derivative).
        Note that this is the forward delta.
        For the spot delta, use `spot_delta`.
        """
        return self._calc_attr(attribute_name="delta")

    def spot_delta(self) -> float:
        """
        Delta discounted for interest rates.
        For the forward delta, use `delta`.
        """
        return self._calc_attr(attribute_name="spot_delta")

    def dual_delta(self) -> float:
        """1st derivative in structure price
        with respect to strike price.
        """
        return self._calc_attr(attribute_name="dual_delta")

    def theta(self) -> float:
        """Rate of change in structure price
        with respect to time (i.e. time decay).
        """
        return self._calc_attr(attribute_name="theta")

    def epsilon(self) -> float:
        """Change in structure price with respect to underlying dividend yield. \n
        Also known as psi."""
        return self._calc_attr(attribute_name="epsilon")

    def rho(self) -> float:
        """Rate of change in structure price
        with respect to the risk-free rate.
        """
        return self._calc_attr(attribute_name="rho")

    def gamma(self) -> float:
        """
        Rate of change in delta with respect to the underlying asset price (2nd derivative).
        """
        return self._calc_attr(attribute_name="gamma")

    def dual_gamma(self) -> float:
        """
        Rate of change in delta with respect to the strike price (2nd derivative).
        """
        return self._calc_attr(attribute_name="dual_gamma")

    def vega(self) -> float:
        """
        Rate of change in structure price with respect to the volatility of the asset.
        """
        return self._calc_attr(attribute_name="vega")

    def lambda_greek(self) -> float:
        """Percentage change in structure price per %
        change in asset price. Also called gearing.
        """
        return self._calc_attr(attribute_name="lambda_greek")

    def vanna(self) -> float:
        """Sensitivity of delta with respect to change in volatility."""
        return self._calc_attr(attribute_name="vanna")

    def charm(self) -> float:
        """Rate of change of delta over time (also known as delta decay)."""
        return self._calc_attr(attribute_name="charm")

    def vomma(self) -> float:
        """2nd order sensitivity to volatility."""
        return self._calc_attr(attribute_name="vomma")

    def veta(self) -> float:
        """Rate of change in `vega` with respect to time."""
        return self._calc_attr(attribute_name="veta")

    def phi(self) -> float:
        """2nd order partial derivative with respect to strike price. \n
        Phi is used in the Breeden-Litzenberger formula. \n
        Breeden-Litzenberger uses quoted option prices
        to estimate risk-neutral probabilities.
        """
        return self._calc_attr(attribute_name="phi")

    def speed(self) -> float:
        """Rate of change in Gamma with respect to change in the underlying asset price."""
        return self._calc_attr(attribute_name="speed")

    def zomma(self) -> float:
        """Rate of change of gamma with respect to changes in volatility."""
        return self._calc_attr(attribute_name="zomma")

    def color(self) -> float:
        """Rate of change of gamma over time."""
        return self._calc_attr(attribute_name="color")

    def ultima(self) -> float:
        """Sensitivity of vomma with respect to change in volatility.
        3rd order derivative of structure price to volatility.
        """
        return self._calc_attr(attribute_name="ultima")

    def alpha(self) -> float:
        """Theta to gamma ratio. Also called "gamma rent".
        More info: "Dynamic Hedging" by Nassim Taleb, p. 178-181.
        """
        return self._calc_attr(attribute_name="alpha")

    def get_core_greeks(self) -> Dict[str, float]:
        """
        Get the top 5 most well known Greeks for the compound.
        1. Delta
        2. Gamma
        3. Vega
        4. Theta
        5. Rho
        """
        return {
            "delta": self.delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "rho": self.rho(),
        }

    def get_all_greeks(self) -> Dict[str, float]:
        """Retrieve all Greeks for the compound
        implemented as a dictionary."""
        return {
            "delta": self.delta(),
            "spot_delta": self.spot_delta(),
            "gamma": self.gamma(),
            "vega": self.vega(),
            "theta": self.theta(),
            "epsilon": self.epsilon(),
            "rho": self.rho(),
            "lambda_greek": self.lambda_greek(),
            "vanna": self.vanna(),
            "charm": self.charm(),
            "vomma": self.vomma(),
            "veta": self.veta(),
            "phi": self.phi(),
            "speed": self.speed(),
            "zomma": self.zomma(),
            "color": self.color(),
            "ultima": self.ultima(),
            "dual_delta": self.dual_delta(),
            "dual_gamma": self.dual_gamma(),
            "alpha": self.alpha(),
        }



from . import BlackScholesCall, BlackScholesPut
from .base import BlackScholesStructureBase


class BlackScholesStraddleLong(BlackScholesStructureBase):
    """
    Create long straddle option structure. \n
    - Long Straddle -> Put(K) + Call(K)

    :param S: Price of underlying asset \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """

    def __init__(
        self,
        S: float,
        K: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        self.call1 = BlackScholesCall(S=S, K=K, T=T, r=r, sigma=sigma, q=q)
        self.put1 = BlackScholesPut(S=S, K=K, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from a put and call option into a long straddle. \n
        All greeks and price are combined in the same way. \n

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to long straddle.
        """
        put_attr = getattr(self.put1, attribute_name)
        call_attr = getattr(self.call1, attribute_name)
        return put_attr() + call_attr()


class BlackScholesStraddleShort(BlackScholesStructureBase):
    """
    Create straddle option structure. \n
    - Short Straddle -> -Put(K) - Call(K)

    :param S: Price of underlying asset \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(
        self,
        S: float,
        K: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        self.call1 = BlackScholesCall(S=S, K=K, T=T, r=r, sigma=sigma, q=q)
        self.put1 = BlackScholesPut(S=S, K=K, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from a put and call option into a short straddle. \n
        All greeks and price are combined in the same way. \n

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to short straddle.
        """
        put_attr = getattr(self.put1, attribute_name)
        call_attr = getattr(self.call1, attribute_name)
        return -put_attr() - call_attr()



from .. import BlackScholesStraddleLong, BlackScholesStraddleShort

# Test parameters
test_S = 55.0  # Asset price of 55
test_K = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesStraddleLong:
    def test_individual_methods(self):
        straddle = BlackScholesStraddleLong(test_S, test_K, test_T, test_r, test_sigma)
        test_methods = list(straddle.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Long straddle = Put1 + Call1
        for attr in test_methods:
            assert (
                getattr(straddle, attr)()
                == getattr(straddle.put1, attr)() + getattr(straddle.call1, attr)()
            )


class TestBlackScholesStraddleShort:
    def test_individual_methods(self):
        straddle = BlackScholesStraddleShort(test_S, test_K, test_T, test_r, test_sigma)
        test_methods = list(straddle.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Short straddle = - Put1 - Call1
        for attr in test_methods:
            assert (
                getattr(straddle, attr)()
                == -getattr(straddle.put1, attr)() - getattr(straddle.call1, attr)()
            )



from . import BlackScholesCall, BlackScholesPut
from .base import BlackScholesStructureBase


class BlackScholesStrangleLong(BlackScholesStructureBase):
    """
    Create long strangle option structure. \n
    - Long strangle -> Put(K1) + Call(K2)

    :param S: Price of underlying asset \n
    :param K1: Strike price for put \n
    :param K2: Strike price for call \n
    It must hold that K1 < K2. \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield) \n
    """

    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 < K2
        ), f"""1st strike price should be smaller than 2nd.
        Got K1={K1}, which is not smaller than K2={K2}.
        """
        self.put1 = BlackScholesPut(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.call1 = BlackScholesCall(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from a put and call option into a long strangle. \n
        All greeks and price are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to long strangle.
        """
        put_attr = getattr(self.put1, attribute_name)
        call_attr = getattr(self.call1, attribute_name)
        return put_attr() + call_attr()


class BlackScholesStrangleShort(BlackScholesStructureBase):
    """
    Create short strangle option structure. \n
    - Short strangle -> -Put(K1) - Call(K2)

    :param S: Price of underlying asset \n
    :param K1: Strike price for put \n
    :param K2: Strike price for call \n
    It must hold that K1 < K2. \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 < K2
        ), f"""1st strike price should be smaller than 2nd.
        Got K1={K1}, which is not smaller than K2={K2}.
        """
        self.put1 = BlackScholesPut(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.call1 = BlackScholesCall(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        super().__init__()

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from a put and call option into a short strangle. \n
        All greeks and price are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to short strangle.
        """
        put_attr = getattr(self.put1, attribute_name)
        call_attr = getattr(self.call1, attribute_name)
        return -put_attr() - call_attr()



import numpy as np
import pytest
from scipy.stats import norm

from blackscholes.base import Black76Base, BlackScholesBase, StandardNormalMixin

# Test parameters
test_S = 55.0  # Asset price of 55
test_K = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestStandardNormalMixIn:
    mix = StandardNormalMixin()
    # PDF and CDF in mixin should be similar to SciPy's standard normal distribution
    for n in range(-10, 10, 2):
        np.testing.assert_almost_equal(norm.pdf(n), mix._pdf(n), decimal=5)
        np.testing.assert_almost_equal(norm.cdf(n), mix._cdf(n), decimal=5)


class BlackScholesMeta(BlackScholesBase):
    """Dummy class for testing Black Scholes base methods."""

    def __init__(
        self, S: float, K: float, T: float, r: float, sigma: float, q: float = 0.0
    ):
        super().__init__(S=S, K=K, T=T, r=r, sigma=sigma, q=q)

    def price(self):
        ...

    def in_the_money(self):
        ...

    def delta(self):
        ...

    def spot_delta(self):
        ...

    def dual_delta(self):
        ...

    def theta(self):
        ...

    def epsilon(self):
        ...

    def rho(self):
        ...

    def charm(self):
        ...


class TestBlackScholesBase:
    meta = BlackScholesMeta(S=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_arg_assert(self):
        # Should not be able to initialize with invalid values for S, K, T, or sigma.
        with pytest.raises(AssertionError):
            BlackScholesMeta(
                S=0.0,
                K=test_K,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
            BlackScholesMeta(
                S=test_S,
                K=0.0,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
            BlackScholesMeta(
                S=test_S,
                K=test_K,
                T=0.0,
                r=test_r,
                sigma=test_sigma,
            )
            BlackScholesMeta(
                S=test_S,
                K=test_K,
                T=test_T,
                r=test_r,
                sigma=0.0,
            )

        # Initializing with negative r (interest rate) is possible.
        BlackScholesMeta(
            S=test_S,
            K=test_K,
            T=test_T,
            r=-test_r,
            sigma=test_sigma,
        )

    def test_d(self):
        # d1 and d2 should be accurate up to at least 6 decimals
        np.testing.assert_almost_equal(self.meta._d1, 0.7270678653621663, decimal=6)
        np.testing.assert_almost_equal(self.meta._d2, 0.5770678653621663, decimal=6)

    def test_gamma(self):
        gamma = self.meta.gamma()
        np.testing.assert_almost_equal(gamma, 0.03712496688031454, decimal=6)

    def test_dual_gamma(self):
        dual_gamma = self.meta.dual_gamma()
        np.testing.assert_almost_equal(dual_gamma, 0.0449212099251806, decimal=6)

    def test_vega(self):
        vega = self.meta.vega()
        np.testing.assert_almost_equal(vega, 16.84545372194272, decimal=6)

    def test_vanna(self):
        vanna = self.meta.vanna()
        np.testing.assert_almost_equal(vanna, -1.178299396409533, decimal=6)

    def test_vomma(self):
        vomma = self.meta.vomma()
        np.testing.assert_almost_equal(vomma, 47.11869947977544, decimal=6)

    def test_veta(self):
        veta = self.meta.veta()
        np.testing.assert_almost_equal(veta, 11.752499520643353, decimal=6)

    def test_phi(self):
        phi = self.meta.phi()
        np.testing.assert_almost_equal(phi, 0.04492120992518061, decimal=6)

    def test_zomma(self):
        zomma = self.meta.zomma()
        np.testing.assert_almost_equal(zomma, -0.14365691533482322, decimal=6)

    def test_speed(self):
        speed = self.meta.speed()
        np.testing.assert_almost_equal(speed, -0.003946801873134375, decimal=6)

    def test_color(self):
        color = self.meta.color()
        np.testing.assert_almost_equal(color, -0.011224141490466934, decimal=6)

    def test_ultima(self):
        ultima = self.meta.ultima()
        np.testing.assert_almost_equal(ultima, -827.4229433648609, decimal=6)


class Black76Meta(Black76Base):
    """Dummy class for testing Black76 base methods."""

    def __init__(self, F: float, K: float, T: float, r: float, sigma: float):
        super().__init__(F=F, K=K, T=T, r=r, sigma=sigma)

    def price(self):
        ...

    def delta(self):
        ...

    def theta(self):
        ...

    def rho(self):
        ...


class TestBlack76Base:
    meta = Black76Meta(F=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_arg_assert(self):
        # Should not be able to initialize if F, K, T, or sigma is negative.
        with pytest.raises(AssertionError):
            Black76Meta(
                F=-test_S,
                K=test_K,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
            Black76Meta(
                F=test_S,
                K=-test_K,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
            Black76Meta(
                F=-test_S,
                K=test_K,
                T=-test_T,
                r=test_r,
                sigma=test_sigma,
            )
            Black76Meta(
                F=test_S,
                K=test_K,
                T=test_T,
                r=test_r,
                sigma=-test_sigma,
            )

        # Initializing with negative r (interest rate) is possible.
        Black76Meta(
            F=test_S,
            K=test_K,
            T=test_T,
            r=-test_r,
            sigma=test_sigma,
        )

    def test_d(self):
        # d1 and d2 should be accurate up to at least 6 decimals
        np.testing.assert_almost_equal(self.meta._d1, 0.7104011986954996, decimal=6)
        np.testing.assert_almost_equal(self.meta._d2, 0.5604011986954995, decimal=6)

    def test_gamma(self):
        gamma = self.meta.gamma()
        np.testing.assert_almost_equal(gamma, 0.03747854417414418, decimal=5)

    def test_vega(self):
        vega = self.meta.vega()
        np.testing.assert_almost_equal(vega, 17.00588941901792, decimal=5)

    def test_vanna(self):
        vanna = self.meta.vanna()
        np.testing.assert_almost_equal(vanna, -1.1551661594303946, decimal=5)

    def test_vomma(self):
        vomma = self.meta.vomma()
        np.testing.assert_almost_equal(vomma, 45.13472833935059, decimal=5)



import pytest

from .. import BlackScholesIronCondorLong, BlackScholesIronCondorShort

# Test parameters
test_S = 55.0  # Asset price of 55
test_K1 = 20.0  # Strike price of 20
test_K2 = 25.0  # Strike price of 25
test_K3 = 45.0  # Strike price of 45
test_K4 = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesIronCondorLong:
    def test_init(self):
        # Assert K1 < K2 < K3 < K4
        with pytest.raises(AssertionError):
            BlackScholesIronCondorLong(
                S=test_S,
                K1=60,
                K2=50,
                K3=40,
                K4=30,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronCondorLong(
                S=test_S,
                K1=41,
                K2=40,
                K3=50,
                K4=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        # Assert K2 - K1 = K3 - K2 (symmetry)
        with pytest.raises(AssertionError):
            BlackScholesIronCondorLong(
                S=test_S,
                K1=20,
                K2=25,
                K3=45,
                K4=61,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronCondorLong(
                S=test_S,
                K1=20,
                K2=26,
                K3=45,
                K4=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )

    def test_individual_methods(self):
        iron_condor = BlackScholesIronCondorLong(
            test_S, test_K1, test_K2, test_K3, test_K4, test_T, test_r, test_sigma
        )
        test_methods = list(iron_condor.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Long iron condor = -Put1 + Put2 + Call1 - Call2
        for attr in test_methods:
            assert (
                getattr(iron_condor, attr)()
                == -getattr(iron_condor.put1, attr)()
                + getattr(iron_condor.put2, attr)()
                + getattr(iron_condor.call1, attr)()
                - getattr(iron_condor.call2, attr)()
            )


class TestBlackScholesIronCondorShort:
    def test_init(self):
        # Assert K1 < K2 < K3 < K4
        with pytest.raises(AssertionError):
            BlackScholesIronCondorShort(
                S=test_S,
                K1=60,
                K2=50,
                K3=40,
                K4=30,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronCondorShort(
                S=test_S,
                K1=41,
                K2=40,
                K3=50,
                K4=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        # Assert K2 - K1 = K3 - K2 (symmetry)
        with pytest.raises(AssertionError):
            BlackScholesIronCondorShort(
                S=test_S,
                K1=20,
                K2=25,
                K3=45,
                K4=61,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronCondorShort(
                S=test_S,
                K1=20,
                K2=26,
                K3=45,
                K4=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )

    def test_individual_methods(self):
        iron_condor = BlackScholesIronCondorShort(
            test_S, test_K1, test_K2, test_K3, test_K4, test_T, test_r, test_sigma
        )
        test_methods = list(iron_condor.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Short iron condor = Put1 - Put2 - Call1 + Call2
        for attr in test_methods:
            assert (
                getattr(iron_condor, attr)()
                == getattr(iron_condor.put1, attr)()
                - getattr(iron_condor.put2, attr)()
                - getattr(iron_condor.call1, attr)()
                + getattr(iron_condor.call2, attr)()
            )



import numpy as np

from blackscholes import Black76Call, Black76Put, BlackScholesCall, BlackScholesPut

# Test parameters
test_S = 55.0  # Asset price of 55
test_K = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesPut:
    put = BlackScholesPut(S=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_price(self):
        put_price = self.put.price()
        np.testing.assert_almost_equal(put_price, 1.214564, decimal=4)

    def test_delta(self):
        put_delta = self.put.delta()
        np.testing.assert_almost_equal(put_delta, -0.233592191490538, decimal=6)

        # Due to put-call parity, Call delta - Put delta should be 1.
        call = BlackScholesCall(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(call.delta() - put_delta, 1.0, decimal=5)

    def test_spot_delta(self):
        put_spot_delta = self.put.spot_delta()
        np.testing.assert_almost_equal(put_spot_delta, -0.23417690255355628, decimal=6)

        # Due to put-call parity, Call spot delta + Put spot delta should be 1.
        call = BlackScholesCall(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(
            call.spot_delta() - put_spot_delta, 1.0, decimal=2
        )

    def test_dual_delta(self):
        put_delta = self.put.dual_delta()
        assert 0.0 < put_delta < 1.0
        np.testing.assert_almost_equal(put_delta, 0.2812428189591384, decimal=6)

        # Due to put-call parity, Call dual delta + Put dual delta should be 1.
        call = BlackScholesCall(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(call.dual_delta() + put_delta, 1.0, decimal=2)

    def test_theta(self):
        put_theta = self.put.theta()
        np.testing.assert_almost_equal(put_theta, -1.2282536767758119, decimal=6)

    def test_epsilon(self):
        call_epsilon = self.put.epsilon()
        np.testing.assert_almost_equal(call_epsilon, 12.84757053197959, decimal=6)

    def test_rho(self):
        put_rho = self.put.rho()
        np.testing.assert_almost_equal(put_rho, -14.062140947956918, decimal=6)

    def test_charm(self):
        charm = self.put.charm()
        np.testing.assert_almost_equal(charm, 0.0832677717846717, decimal=6)

    def test_get_core_greeks(self):
        core_greeks = self.put.get_core_greeks()
        expected_result = {
            "delta": -0.233592191490538,
            "gamma": 0.03712496688031454,
            "vega": 16.84545372194272,
            "theta": -1.2282536767758119,
            "rho": -14.062140947956918,
        }
        assert set(core_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                core_greeks[key], expected_result[key], decimal=5
            )

    def test_get_itm_proxies(self):
        itm_proxies = self.put.get_itm_proxies()
        expected_result = {
            "in_the_money": 0.2819468056232066,
            "dual_delta": 0.2812428189591384,
        }
        assert set(itm_proxies.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                itm_proxies[key], expected_result[key], decimal=5
            )

    def test_in_the_money(self):
        itm_prob = self.put.in_the_money()
        assert 0.0 < itm_prob < 1.0
        np.testing.assert_almost_equal(itm_prob, 0.2819468056232066, decimal=6)

    def test_all_greeks(self):
        all_greeks = self.put.get_all_greeks()
        expected_result = {
            "delta": -0.233592191490538,
            "spot_delta": -0.23417690255355628,
            "gamma": 0.03712496688031454,
            "vega": 16.84545372194272,
            "theta": -1.2282536767758119,
            "epsilon": 12.84757053197959,
            "rho": -14.062140947956918,
            "lambda_greek": -10.57787211261979,
            "vanna": -1.178299396409533,
            "charm": 0.0832677717846717,
            "vomma": 47.11869947977544,
            "veta": 11.752499520643353,
            "phi": 0.04492120992518061,
            "speed": -0.003946801873134375,
            "zomma": -0.14365691533482322,
            "color": -0.011224141490466934,
            "ultima": -827.4229433648609,
            "dual_delta": 0.2812428189591384,
            "dual_gamma": 0.0449212099251806,
            "alpha": 33.084302745675686,
        }
        assert set(all_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                all_greeks[key], expected_result[key], decimal=5
            )


class TestBlack76Put:
    put = Black76Put(F=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_price(self):
        price = self.put.price()
        np.testing.assert_almost_equal(price, 1.2470010007171901, decimal=6)

    def test_delta(self):
        delta = self.put.delta()
        np.testing.assert_almost_equal(delta, -0.23813161620464113, decimal=5)

        # Due to put-call parity, Call delta - Put delta should be 1.
        call = Black76Call(
            F=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(call.delta() - delta, 1.0, decimal=2)

    def test_theta(self):
        theta = self.put.theta()
        np.testing.assert_almost_equal(theta, -1.272324203924551, decimal=5)

    def test_rho(self):
        rho = self.put.rho()
        np.testing.assert_almost_equal(rho, -1.2470010007171901, decimal=5)

    def test_get_core_greeks(self):
        core_greeks = self.put.get_core_greeks()
        expected_result = {
            "delta": -0.23813161620464113,
            "gamma": 0.03747854417414418,
            "vega": 17.00588941901792,
            "theta": -1.272324203924551,
            "rho": -1.2470010007171901,
        }
        assert set(core_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                core_greeks[key], expected_result[key], decimal=5
            )

    def test_all_greeks(self):
        all_greeks = self.put.get_all_greeks()
        expected_result = {
            "delta": -0.23813161620464113,
            "gamma": 0.03747854417414418,
            "vega": 17.00588941901792,
            "theta": -1.272324203924551,
            "rho": -1.2470010007171901,
            "vanna": -1.1551661594303946,
            "vomma": 45.13472833935059,
            "alpha": 33.94806810170172,
        }

        assert set(all_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                all_greeks[key], expected_result[key], decimal=5
            )



from . import BlackScholesCall, BlackScholesPut
from .base import BlackScholesStructureBase


class BlackScholesButterflyLong(BlackScholesStructureBase):
    """
    Create long butterfly option structure.
    - Long butterfly -> Call(K1) - 2 * Call(K2) + Call(K3)

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st option \n
    :param K2: Strike price for 2nd option \n
    :param K3: Strike price for 3rd option \n
    It must hold that K1 < K2 < K3. \n
    Additionally, it must hold that K2 - K1 = K3 - K2 \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        K3: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 < K2 < K3
        ), f"""It must hold that K1 < K2 < K3.
                        Got {K1}, {K2} and {K3}.
                        """
        assert (
            K2 - K1 == K3 - K2
        ), f"""Strike price must be symmetric, so K2 - K1 = K3 - K2.
                        Got {K2}-{K1} != {K3}-{K2}.
                        """
        super().__init__()
        self.call1 = BlackScholesCall(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.call2 = BlackScholesCall(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        self.call3 = BlackScholesCall(S=S, K=K3, T=T, r=r, sigma=sigma, q=q)

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from three call options into a long call butterfly. \n
        All greeks and price are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to long call butterfly.
        """
        call_attr1 = getattr(self.call1, attribute_name)
        call_attr2 = getattr(self.call2, attribute_name)
        call_attr3 = getattr(self.call3, attribute_name)
        return call_attr1() - 2 * call_attr2() + call_attr3()


class BlackScholesButterflyShort(BlackScholesStructureBase):
    """
    Create short butterfly option structure. \n
    - Short butterfly -> -Put(K1) + 2 * Put(K2) - Put(K3)

    :param S: Price of underlying asset \n
    :param K1: Strike price for 1st option \n
    :param K2: Strike price for 2nd option \n
    :param K3: Strike price for 3rd option \n
    It must hold that K1 < K2 < K3. \n
    Additionally, it must hold that K2 - K1 = K3 - K2 \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(
        self,
        S: float,
        K1: float,
        K2: float,
        K3: float,
        T: float,
        r: float,
        sigma: float,
        q: float = 0.0,
    ):
        assert (
            K1 < K2 < K3
        ), f"""It must hold that K1 < K2 < K3.
                        Got {K1}, {K2} and {K3}.
                        """
        assert (
            K2 - K1 == K3 - K2
        ), f"""Strike price must be symmetric, so K2 - K1 = K3 - K2.
                        Got {K2}-{K1} != {K3}-{K2}.
                        """
        super().__init__()
        self.put1 = BlackScholesPut(S=S, K=K1, T=T, r=r, sigma=sigma, q=q)
        self.put2 = BlackScholesPut(S=S, K=K2, T=T, r=r, sigma=sigma, q=q)
        self.put3 = BlackScholesPut(S=S, K=K3, T=T, r=r, sigma=sigma, q=q)

    def _calc_attr(self, attribute_name: str) -> float:
        """
        Combines attributes from three put options into a short put butterfly. \n
        All greeks and price are combined in the same way.

        :param attribute_name: String name of option attribute
        pointing to a method that can be called on
        BlackScholesCall and BlackScholesPut.

        :return: Combined value according to short put butterfly.
        """
        put_attr1 = getattr(self.put1, attribute_name)
        put_attr2 = getattr(self.put2, attribute_name)
        put_attr3 = getattr(self.put3, attribute_name)
        return -put_attr1() + 2 * put_attr2() - put_attr3()



import pytest

from .. import (BlackScholesBullSpread, BlackScholesBearSpread, 
                BlackScholesCalendarCallSpread, BlackScholesCalendarPutSpread)

# Test parameters
test_S = 55.0  # Asset price of 55
test_K1 = 40.0  # Strike price of 40
test_K2 = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_T2 = 0.5  # 6 months to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesBullSpread:
    def test_init(self):
        # Assert K1 < K2
        with pytest.raises(AssertionError):
            BlackScholesBullSpread(
                S=test_S, K1=50, K2=45, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        spread = BlackScholesBullSpread(
            test_S, test_K1, test_K2, test_T, test_r, test_sigma
        )
        test_methods = list(spread.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Bull spread = Call1 - Call2
        for attr in test_methods:
            assert (
                getattr(spread, attr)()
                == getattr(spread.call1, attr)() - getattr(spread.call2, attr)()
            )


class TestBlackScholesBearSpread:
    def test_init(self):
        # Assert K1 > K2
        with pytest.raises(AssertionError):
            BlackScholesBearSpread(
                S=test_S, K1=45, K2=50, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        spread = BlackScholesBearSpread(
            test_S, test_K2, test_K1, test_T, test_r, test_sigma
        )
        test_methods = list(spread.put1.get_all_greeks().keys()) + [
            "price",
        ]
        # Bear spread = Put1 - Put2
        for attr in test_methods:
            assert (
                getattr(spread, attr)()
                == getattr(spread.put1, attr)() - getattr(spread.put2, attr)()
            )

class TestBlackScholesCalendarCallSpread:
    def test_init(self):
        # Assert T1 > T2
        with pytest.raises(AssertionError):
            BlackScholesCalendarCallSpread(
                S=test_S, K1=test_K1, K2=test_K2, T1=1., T2=1.5, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        spread = BlackScholesCalendarCallSpread(
            test_S, test_K1, test_K2, test_T, test_T2, test_r, test_sigma
        )
        test_methods = list(spread.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Calendar Call Spread = Call1 - Call2
        for attr in test_methods:
            assert (
                getattr(spread, attr)()
                == getattr(spread.call1, attr)() - getattr(spread.call2, attr)()
            )

class TestBlackScholesCalendarPutSpread:
    def test_init(self):
        # Assert T1 > T2
        with pytest.raises(AssertionError):
            BlackScholesCalendarPutSpread(
                S=test_S, K1=test_K1, K2=test_K2, T1=1., T2=1.5, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        spread = BlackScholesCalendarPutSpread(
            test_S, test_K1, test_K2, test_T, test_T2, test_r, test_sigma
        )
        test_methods = list(spread.put1.get_all_greeks().keys()) + [
            "price",
        ]
        # Calendar Put Spread = Put1 - Put2
        for attr in test_methods:
            assert (
                getattr(spread, attr)()
                == getattr(spread.put1, attr)() - getattr(spread.put2, attr)()
            )



from math import exp, sqrt

from .base import Black76Base, BlackScholesBase


class BlackScholesCall(BlackScholesBase):
    """
    Class to calculate (European) call option prices
    and Greeks with the Black-Scholes-Merton formula.

    :param S: Price of underlying asset \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    :param q: Annual dividend yield (0.05 indicates 5% yield)
    """

    def __init__(
        self, S: float, K: float, T: float, r: float, sigma: float, q: float = 0.0
    ):
        super().__init__(S=S, K=K, T=T, r=r, sigma=sigma, q=q)

    def price(self) -> float:
        """Fair value of Black-Scholes call option."""
        return (
            self.S * exp(-self.q * self.T) * self._cdf(self._d1)
            - self._cdf(self._d2) * exp(-self.r * self.T) * self.K
        )

    def delta(self) -> float:
        """Rate of change in option price
        with respect to the forward price (1st derivative).
        Note that this is the forward delta.
        For the spot delta, use `spot_delta`.
        """
        return exp(-self.q * self.T) * self._cdf(self._d1)

    def spot_delta(self) -> float:
        """
        Delta discounted for interest rates.
        For the forward delta, use `delta`.
        """
        return exp((self.r - self.q) * self.T) * self._cdf(self._d1)

    def dual_delta(self) -> float:
        """1st derivative in option price
        with respect to strike price.
        """
        return exp(-self.r * self.T) * self._cdf(self._d2)

    def theta(self) -> float:
        """Rate of change in option price
        with respect to time (i.e. time decay).
        """
        return (
            (-exp(-self.q * self.T) * self.S * self._pdf(self._d1) * self.sigma)
            / (2 * sqrt(self.T))
            - (self.r * self.K * exp(-self.r * self.T) * self._cdf(self._d2))
            + self.q * self.S * exp(-self.q * self.T) * self._cdf(self._d1)
        )

    def rho(self) -> float:
        """Rate of change in option price
        with respect to the risk-free rate.
        """
        return self.K * self.T * exp(-self.r * self.T) * self._cdf(self._d2)

    def epsilon(self) -> float:
        """Change in option price with respect to underlying dividend yield. \n
        Also known as psi."""
        return -self.S * self.T * exp(-self.q * self.T) * self._cdf(self._d1)

    def charm(self) -> float:
        """Rate of change of delta over time (also known as delta decay)."""
        return self.q * exp(-self.q * self.T) * self._cdf(self._d1) - exp(
            -self.q * self.T
        ) * self._pdf(self._d1) * (
            2.0 * (self.r - self.q) * self.T - self._d2 * self.sigma * sqrt(self.T)
        ) / (
            2.0 * self.T * self.sigma * sqrt(self.T)
        )

    def in_the_money(self) -> float:
        """Naive Probability that call option will be in the money at maturity."""
        return self._cdf(self._d2)


class Black76Call(Black76Base):
    """
    Class to calculate (European) call option prices
    and Greeks with the Black-76 formula.

    :param F: Price of underlying futures contract \n
    :param K: Strike price \n
    :param T: Time till expiration in years (1/12 indicates 1 month) \n
    :param r: Risk-free interest rate (0.05 indicates 5%) \n
    :param sigma: Volatility (standard deviation) of stock (0.15 indicates 15%) \n
    """

    def __init__(self, F: float, K: float, T: float, r: float, sigma: float):
        super().__init__(F=F, K=K, T=T, r=r, sigma=sigma)

    def price(self) -> float:
        """Fair value of a Black-76 call option."""
        return exp(-self.r * self.T) * (
            self.F * self._cdf(self._d1) - self.K * self._cdf(self._d2)
        )

    def delta(self) -> float:
        """Rate of change in option price
        with respect to the underlying futures price (1st derivative).
        Proxy for probability of the option expiring in the money.
        """
        return exp(-self.r * self.T) * self._cdf(self._d1)

    def theta(self) -> float:
        """Rate of change in option price
        with respect to time (i.e. time decay).
        """
        return (
            -self.F
            * exp(-self.r * self.T)
            * self._pdf(self._d1)
            * self.sigma
            / (2 * sqrt(self.T))
            - self.r * self.K * exp(-self.r * self.T) * self._cdf(self._d2)
            + self.r * self.F * exp(-self.r * self.T) * self._cdf(self._d1)
        )

    def rho(self) -> float:
        """Rate of change in option price
        with respect to the risk-free rate.
        """
        return (
            -self.T
            * exp(-self.r * self.T)
            * (self.F * self._cdf(self._d1) - self.K * self._cdf(self._d2))
        )



from .call import Black76Call, BlackScholesCall
from .put import Black76Put, BlackScholesPut
from .straddle import BlackScholesStraddleLong, BlackScholesStraddleShort
from .strangle import BlackScholesStrangleLong, BlackScholesStrangleShort
from .butterfly import BlackScholesButterflyLong, BlackScholesButterflyShort
from .iron_condor import BlackScholesIronCondorLong, BlackScholesIronCondorShort
from .spread import (BlackScholesBullSpread, BlackScholesBearSpread, 
                     BlackScholesCalendarCallSpread, BlackScholesCalendarPutSpread) 
from .iron_butterfly import (
    BlackScholesIronButterflyLong,
    BlackScholesIronButterflyShort,
)

__all__ = [
    "BlackScholesCall",
    "Black76Call",
    "BlackScholesPut",
    "Black76Put",
    "BlackScholesStraddleLong",
    "BlackScholesStraddleShort",
    "BlackScholesStrangleLong",
    "BlackScholesStrangleShort",
    "BlackScholesButterflyLong",
    "BlackScholesButterflyShort",
    "BlackScholesIronCondorLong",
    "BlackScholesIronCondorShort",
    "BlackScholesBullSpread",
    "BlackScholesBearSpread",
    "BlackScholesCalendarCallSpread",
    "BlackScholesCalendarPutSpread",
    "BlackScholesIronButterflyLong",
    "BlackScholesIronButterflyShort",
]



import pytest

from .. import BlackScholesStrangleLong, BlackScholesStrangleShort

# Test parameters
test_S = 55.0  # Asset price of 55
test_K1 = 40.0  # Strike price of 40
test_K2 = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesStrangleLong:
    def test_init(self):
        # Assert K1 < K2
        with pytest.raises(AssertionError):
            BlackScholesStrangleLong(
                S=test_S, K1=50, K2=45, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        strangle = BlackScholesStrangleLong(
            test_S, test_K1, test_K2, test_T, test_r, test_sigma
        )
        test_methods = list(strangle.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Long strangle = Put1 + Call1
        for attr in test_methods:
            assert (
                getattr(strangle, attr)()
                == getattr(strangle.put1, attr)() + getattr(strangle.call1, attr)()
            )


class TestBlackScholesStrangleShort:
    def test_init(self):
        # Assert K1 < K2
        with pytest.raises(AssertionError):
            BlackScholesStrangleShort(
                S=test_S, K1=50, K2=45, T=test_T, r=test_r, sigma=test_sigma
            )

    def test_individual_methods(self):
        strangle = BlackScholesStrangleShort(
            test_S, test_K1, test_K2, test_T, test_r, test_sigma
        )
        test_methods = list(strangle.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Short strangle = -Put1 - Call1
        for attr in test_methods:
            assert (
                getattr(strangle, attr)()
                == -getattr(strangle.put1, attr)() - getattr(strangle.call1, attr)()
            )



import numpy as np

from blackscholes import Black76Call, Black76Put, BlackScholesCall, BlackScholesPut

# Test parameters
test_S = 55.0  # Asset price of 55
test_K = 50.0  # Strike price of 50
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesCall:
    call = BlackScholesCall(S=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_price(self):
        call_price = self.call.price()
        np.testing.assert_almost_equal(call_price, 6.339408, decimal=4)

    def test_delta(self):
        call_delta = self.call.delta()
        np.testing.assert_almost_equal(call_delta, 0.766407808509462, decimal=6)

        # Due to put-call parity, Call delta - Put delta should be 1.
        put = BlackScholesPut(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(call_delta - put.delta(), 1.0, decimal=5)

    def test_spot_delta(self):
        call_spot_delta = self.call.spot_delta()

        np.testing.assert_almost_equal(call_spot_delta, 0.7683262250522389, decimal=6)

        # Due to put-call parity, Call spot delta + Put spot delta should be 1.
        put = BlackScholesPut(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(
            call_spot_delta - put.spot_delta(), 1.0, decimal=2
        )

    def test_dual_delta(self):
        call_delta = self.call.dual_delta()
        assert 0.0 < call_delta < 1.0
        np.testing.assert_almost_equal(call_delta, 0.7162603034383217, decimal=6)

        # Due to put-call parity, Put dual delta + call dual delta should be 1.
        put = BlackScholesPut(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(put.dual_delta() + call_delta, 1.0, decimal=2)

    def test_theta(self):
        call_theta = self.call.theta()
        np.testing.assert_almost_equal(call_theta, -1.3529415670754943, decimal=6)

    def test_epsilon(self):
        call_epsilon = self.call.epsilon()
        np.testing.assert_almost_equal(call_epsilon, -42.15242946802041, decimal=6)

    def test_rho(self):
        call_rho = self.call.rho()
        np.testing.assert_almost_equal(call_rho, 35.813015171916085, decimal=6)

    def test_get_core_greeks(self):
        core_greeks = self.call.get_core_greeks()
        expected_result = {
            "delta": 0.766407808509462,
            "gamma": 0.03712496688031454,
            "vega": 16.84545372194272,
            "theta": -1.3529415670754943,
            "rho": 35.813015171916085,
        }
        assert set(core_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                core_greeks[key], expected_result[key], decimal=5
            )

    def test_get_itm_proxies(self):
        itm_proxies = self.call.get_itm_proxies()
        expected_result = {
            "in_the_money": 0.7180531943767934,
            "dual_delta": 0.7162603034383217,
        }
        assert set(itm_proxies.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                itm_proxies[key], expected_result[key], decimal=5
            )

    def test_in_the_money(self):
        itm_prob_call = self.call.in_the_money()
        assert 0.0 < itm_prob_call < 1.0
        np.testing.assert_almost_equal(itm_prob_call, 0.7180531943767934, decimal=6)

        # Due to put-call parity, Call itm + Put itm should be 1.
        put = BlackScholesPut(
            S=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        assert itm_prob_call + put.in_the_money() == 1.0

    def test_lambda(self):
        lambda_greek = self.call.lambda_greek()
        np.testing.assert_almost_equal(lambda_greek, 6.6492624553539335, decimal=6)

    def test_charm(self):
        charm = self.call.charm()
        np.testing.assert_almost_equal(charm, 0.0832677717846717, decimal=6)

    def test_all_greeks(self):
        all_greeks = self.call.get_all_greeks()
        expected_result = {
            "delta": 0.766407808509462,
            "spot_delta": 0.7683262250522389,
            "gamma": 0.03712496688031454,
            "vega": 16.84545372194272,
            "theta": -1.3529415670754943,
            "epsilon": -42.15242946802041,
            "rho": 35.813015171916085,
            "lambda_greek": 6.6492624553539255,
            "vanna": -1.178299396409533,
            "charm": 0.0832677717846717,
            "vomma": 47.11869947977544,
            "veta": 11.752499520643353,
            "phi": 0.04492120992518061,
            "speed": -0.003946801873134375,
            "zomma": -0.14365691533482322,
            "color": -0.011224141490466934,
            "ultima": -827.4229433648609,
            "dual_delta": 0.7162603034383217,
            "dual_gamma": 0.0449212099251806,
            "alpha": 36.442902023166184,
        }
        assert set(all_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                all_greeks[key], expected_result[key], decimal=5
            )


class TestBlack76Call:
    call = Black76Call(F=test_S, K=test_K, T=test_T, r=test_r, sigma=test_sigma)

    def test_price(self):
        price = self.call.price()
        np.testing.assert_almost_equal(price, 6.234516612704489, decimal=6)

    def test_delta(self):
        delta = self.call.delta()
        np.testing.assert_almost_equal(delta, 0.7593715061928189, decimal=5)

        # Due to put-call parity, Call delta - Put delta should be 1.
        put = Black76Put(
            F=test_S,
            K=test_K,
            T=test_T,
            r=test_r,
            sigma=test_sigma,
        )
        np.testing.assert_almost_equal(delta - put.delta(), 1.0, decimal=2)

    def test_theta(self):
        theta = self.call.theta()
        np.testing.assert_almost_equal(theta, -1.2598554148945826, decimal=5)

    def test_rho(self):
        rho = self.call.rho()
        np.testing.assert_almost_equal(rho, -6.234516612704489, decimal=5)

    def test_get_core_greeks(self):
        core_greeks = self.call.get_core_greeks()
        expected_result = {
            "delta": 0.7593715061928189,
            "gamma": 0.03747854417414418,
            "vega": 17.00588941901792,
            "theta": -1.2598554148945826,
            "rho": -6.234516612704489,
        }

        assert set(core_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                core_greeks[key], expected_result[key], decimal=5
            )

    def test_all_greeks(self):
        all_greeks = self.call.get_all_greeks()
        expected_result = {
            "delta": 0.7593715061928189,
            "gamma": 0.03747854417414418,
            "vega": 17.00588941901792,
            "theta": -1.2598554148945826,
            "rho": -6.234516612704489,
            "vanna": -1.1551661594303946,
            "vomma": 45.13472833935059,
            "alpha": 33.615376718617554,
        }
        assert set(all_greeks.keys()) == set(expected_result.keys())
        for key in expected_result.keys():
            np.testing.assert_almost_equal(
                all_greeks[key], expected_result[key], decimal=5
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blackscholes', 'blackscholes.tests']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'blackscholes',
    'version': '0.1.6',
    'description': 'Black Scholes calculator for Python including all Greeks',
    'long_description': '# blackscholes\n\n![](https://img.shields.io/pypi/dm/blackscholes) | \n![](https://img.shields.io/pypi/pyversions/blackscholes) | \n![](https://img.shields.io/codecov/c/github/carlolepelaars/blackscholes)\n\nA Black-Scholes calculator for Python that includes up to the third-order Greeks.\n\nSupports the Black-Scholes-Merton model, \nBlack-76 model and option structures.\n\n## Installation\n\n`pip install blackscholes`\n\n## Examples\n\n### Input variables\n```python3\nS = 55.0  # Asset price of 55\nK = 50.0  # Strike price of 50\nT = 1.0  # 1 Year to maturity\nr = 0.0025  # 0.25% Risk-free rate\nsigma = 0.15  # 15% Volatility\nq = 0.0 # 0% Annual Dividend Yield\n```\n\n### Call\n\n```python3\nfrom blackscholes import BlackScholesCall\ncall = BlackScholesCall(S=S, K=K, T=T, r=r, sigma=sigma, q=q)\ncall.price()  ## 6.339408\ncall.delta()  ## 0.766407\ncall.spot_delta() ## 0.7683\ncall.charm()  ## 0.083267\n```\n\n### Put\n\n```python3\nfrom blackscholes import BlackScholesPut\nput = BlackScholesPut(S=S, K=K, T=T, r=r, sigma=sigma, q=q)\nput.price()  ## 1.214564\nput.delta()  ## -0.23359\nput.spot_delta() ## -0.23417\nput.charm()  ## 0.083267\n```\n\n### Black-76\n\nThe Black-76 model is often used specifically for options and futures and bonds.\n`blackscholes` also supports this model. To see all available greeks\ncheck out section [4. The Greeks (Black-76)](https://carlolepelaars.github.io/blackscholes/4.the_greeks_black76).\n\n**Call Example**\n```python\nfrom blackscholes import Black76Call\ncall = Black76Call(F=55, K=50, T=1, r=0.0025, sigma=0.15)\ncall.price()  ## 6.2345\ncall.delta()  ## 0.7594\ncall.vomma()  ## 45.1347\n```\n\n**Put Example**\n```python\nfrom blackscholes import Black76Put\nput = Black76Put(F=55, K=50, T=1, r=0.0025, sigma=0.15)\nput.price()  ## 1.2470\nput.delta()  ## -0.2381\nput.vomma()  ## 45.1347\n```\n\n### Structures\n\n`blackscholes` offers the following six option structures:\n- Straddle\n- Strangle\n- Butterfly\n- Iron Condor\n- Spreads\n- Iron Butterfly\n\nAll structures have a long and short version. To learn more\ncheck out section [6. Option Structures](https://carlolepelaars.github.io/blackscholes/6.option_structures).\n\n**Long Straddle Example**\n```python3\nfrom blackscholes import BlackScholesStraddleLong\n\nstraddle = BlackScholesStraddleLong(S=55, K=50, T=1.0,\n                                    r=0.0025, sigma=0.15)\nstraddle.price()  ## 7.5539\nstraddle.delta()  ## 0.5328\n```\n\n## Contributing\n\nWe very much welcome new contributions! Check out the [Github Issues](https://github.com/CarloLepelaars/blackscholes/issues)\nto see what is currently being worked on.\n\nAlso check out [Contributing](https://carlolepelaars.github.io/blackscholes/contributing) in the documentation \nto learn more about \ncontributing to [blackscholes](https://github.com/CarloLepelaars/blackscholes).\n',
    'author': 'CarloLepelaars',
    'author_email': 'info@carlolepelaars.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



import pytest

from .. import BlackScholesIronButterflyLong, BlackScholesIronButterflyShort

# Test parameters
test_S = 25.0  # Asset price of 30
test_K1 = 20.0  # Strike price of 20
test_K2 = 25.0  # Strike price of 25
test_K3 = 30.0  # Strike price of 30
test_T = 1.0  # 1 year to maturity
test_r = 0.0025  # 0.25% risk-free rate
test_sigma = 0.15  # 15% vol


class TestBlackScholesIronButterflyLong:
    def test_init(self):
        # Assert K1 < K2 < K3
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyLong(
                S=test_S,
                K1=60,
                K2=50,
                K3=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyLong(
                S=test_S,
                K1=40,
                K2=40,
                K3=50,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        # Assert equidistance between strike prices
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyLong(
                S=test_S,
                K1=19,
                K2=25,
                K3=35,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyLong(
                S=test_S,
                K1=20,
                K2=25,
                K3=36,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyLong(
                S=test_S,
                K1=19,
                K2=25,
                K3=30,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )

    def test_individual_methods(self):
        iron_butterfly = BlackScholesIronButterflyLong(
            test_S, test_K1, test_K2, test_K3, test_T, test_r, test_sigma
        )
        test_methods = list(iron_butterfly.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Long iron butterfly = -Put1 + Put2 + Call1 - Call2
        for attr in test_methods:
            assert (
                getattr(iron_butterfly, attr)()
                == -getattr(iron_butterfly.put1, attr)()
                + getattr(iron_butterfly.put2, attr)()
                + getattr(iron_butterfly.call1, attr)()
                - getattr(iron_butterfly.call2, attr)()
            )


class TestBlackScholesIronButterflyShort:
    def test_init(self):
        # Assert K1 < K2 < K3
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyShort(
                S=test_S,
                K1=60,
                K2=50,
                K3=60,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyShort(
                S=test_S,
                K1=40,
                K2=40,
                K3=50,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        # Assert equidistance between strike prices
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyShort(
                S=test_S,
                K1=19,
                K2=25,
                K3=35,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyShort(
                S=test_S,
                K1=20,
                K2=25,
                K3=36,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )
        with pytest.raises(AssertionError):
            BlackScholesIronButterflyShort(
                S=test_S,
                K1=19,
                K2=25,
                K3=30,
                T=test_T,
                r=test_r,
                sigma=test_sigma,
            )

    def test_individual_methods(self):
        iron_butterfly = BlackScholesIronButterflyShort(
            test_S, test_K1, test_K2, test_K3, test_T, test_r, test_sigma
        )
        test_methods = list(iron_butterfly.call1.get_all_greeks().keys()) + [
            "price",
        ]
        # Short iron butterfly = Put1 - Put2 - Call1 + Call2
        for attr in test_methods:
            assert (
                getattr(iron_butterfly, attr)()
                == getattr(iron_butterfly.put1, attr)()
                - getattr(iron_butterfly.put2, attr)()
                - getattr(iron_butterfly.call1, attr)()
                + getattr(iron_butterfly.call2, attr)()
            )
